@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General
  * Public License along with this library; if not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -35,10 +35,13 @@
 
 #define Q(a) { a, a, a, a }
 
-static const float BABL_ALPHA_THRESHOLD_FLOAT = (float)BABL_ALPHA_THRESHOLD;
+static const float BABL_ALPHA_FLOOR_FLOAT = (float)BABL_ALPHA_FLOOR;
 
-static long
-conv_rgbaF_linear_rgbAF_linear (const float *src, float *dst, long samples)
+static void
+conv_rgbaF_linear_rgbAF_linear (const Babl  *conversion,
+                                const float *src,
+                                float       *dst,
+                                long         samples)
 {
   long i = 0;
   long remainder;
@@ -51,19 +54,29 @@ conv_rgbaF_linear_rgbAF_linear (const float *src, float *dst, long samples)
 
       for ( ; i < n; i += 2)
         {
+          float alpha0 = ((float *)s)[3];
+          float alpha1 = ((float *)s)[7];
+          float used_alpha0 = babl_epsilon_for_zero_float (alpha0);
+          float used_alpha1 = babl_epsilon_for_zero_float (alpha1);
+
+         {
           __v4sf rbaa0, rbaa1;
         
           __v4sf rgba0 = *s++;
           __v4sf rgba1 = *s++;
 
+
           /* Expand alpha */
-          __v4sf aaaa0 = (__v4sf)_mm_shuffle_epi32((__m128i)rgba0, _MM_SHUFFLE(3, 3, 3, 3));
-          __v4sf aaaa1 = (__v4sf)_mm_shuffle_epi32((__m128i)rgba1, _MM_SHUFFLE(3, 3, 3, 3));
+          __v4sf aaaa0 = (__v4sf)_mm_set1_ps(used_alpha0);
+          __v4sf aaaa1 = (__v4sf)_mm_set1_ps(used_alpha1);
           
           /* Premultiply */
           rgba0 = rgba0 * aaaa0;
           rgba1 = rgba1 * aaaa1;
-          
+    
+          aaaa0 = (__v4sf)_mm_set1_ps(alpha0);
+          aaaa1 = (__v4sf)_mm_set1_ps(alpha1);
+
           /* Shuffle the original alpha value back in */
           rbaa0 = _mm_shuffle_ps(rgba0, aaaa0, _MM_SHUFFLE(0, 0, 2, 0));
           rbaa1 = _mm_shuffle_ps(rgba1, aaaa1, _MM_SHUFFLE(0, 0, 2, 0));
@@ -73,6 +86,7 @@ conv_rgbaF_linear_rgbAF_linear (const float *src, float *dst, long samples)
           
           *d++ = rgba0;
           *d++ = rgba1;
+         }
         }
       _mm_empty ();
     }
@@ -82,21 +96,23 @@ conv_rgbaF_linear_rgbAF_linear (const float *src, float *dst, long samples)
   remainder = samples - i;
   while (remainder--)
   {
-    const float a = src[3];
-    dst[0] = src[0] * a;
-    dst[1] = src[1] * a;
-    dst[2] = src[2] * a;
+    float a = src[3];
+    float used_alpha = babl_epsilon_for_zero_float (a);
+    dst[0] = src[0] * used_alpha;
+    dst[1] = src[1] * used_alpha;
+    dst[2] = src[2] * used_alpha;
     dst[3] = a;
     
     src += 4;
     dst += 4;
   }
-
-  return samples;
 }
 
-static long
-conv_rgbAF_linear_rgbaF_linear_shuffle (const float *src, float *dst, long samples)
+static void
+conv_rgbAF_linear_rgbaF_linear_shuffle (const Babl  *conversion,
+                                        const float *src,  
+                                        float       *dst, 
+                                        long         samples)
 {
   long i = 0;
   long remainder;
@@ -112,16 +128,11 @@ conv_rgbAF_linear_rgbaF_linear_shuffle (const float *src, float *dst, long sampl
           __v4sf pre_rgba0, rgba0, rbaa0, raaaa0;
           
           float alpha0 = ((float *)s)[3];
+          float used_alpha0 = babl_epsilon_for_zero_float (alpha0);
           pre_rgba0 = *s;
           
-          if (alpha0 <= 0.0f)
           {
-            /* Zero RGB */
-            rgba0 = _mm_setzero_ps();
-          }
-          else
-          {
-            float recip0 = 1.0f/alpha0;
+            float recip0 = 1.0f/used_alpha0;
             
             /* Expand reciprocal */
             raaaa0 = _mm_load1_ps(&recip0);
@@ -133,7 +144,7 @@ conv_rgbAF_linear_rgbaF_linear_shuffle (const float *src, float *dst, long sampl
           /* Shuffle the original alpha value back in */
           rbaa0 = _mm_shuffle_ps(rgba0, pre_rgba0, _MM_SHUFFLE(3, 3, 2, 0));
           rgba0 = _mm_shuffle_ps(rgba0, rbaa0, _MM_SHUFFLE(2, 1, 1, 0));
-          
+
           s++;
           *d++ = rgba0;
         }
@@ -159,22 +170,23 @@ conv_rgbAF_linear_rgbaF_linear_shuffle (const float *src, float *dst, long sampl
       src   += 4;
       dst   += 4;
     }
-
-  return samples;
 }
 
-static long
-conv_rgbAF_linear_rgbaF_linear_spin (const float *src, float *dst, long samples)
+static void
+conv_rgbAF_linear_rgbaF_linear_spin (const Babl  *conversion,
+                                     const float *src,  
+                                     float       *dst, 
+                                     long         samples)
 {
   long i = 0;
   long remainder;
-
+  // XXX : not ported to zero preserving alpha transforms
   if (((uintptr_t)src % 16) + ((uintptr_t)dst % 16) == 0)
     {
       const long    n = samples;
       const __v4sf *s = (const __v4sf*) src;
             __v4sf *d = (__v4sf*)dst;
-      const __v4sf zero = _mm_setzero_ps();
+      const __v4sf zero = _mm_set_ss (BABL_ALPHA_FLOOR_FLOAT);
       const __v4sf one = _mm_set_ss(1.0f);
 
       for ( ; i < n; i += 1)
@@ -234,8 +246,6 @@ conv_rgbAF_linear_rgbaF_linear_spin (const float *src, float *dst, long samples)
       src   += 4;
       dst   += 4;
     }
-
-  return samples;
 }
 
 #define splat4f(x) ((__v4sf){x,x,x,x})
@@ -243,8 +253,24 @@ conv_rgbAF_linear_rgbaF_linear_spin (const float *src, float *dst, long samples)
 #define FLT_ONE 0x3f800000 // ((union {float f; int i;}){1.0f}).i
 #define FLT_MANTISSA (1<<23)
 
+static inline float
+sse_max_component (__v4sf x) {
+  __v4sf s;
+  __v4sf m;
+
+  /* m = [max (x[3], x[1]), max (x[2], x[0])] */
+  s = (__v4sf) _mm_shuffle_epi32 ((__m128i) x, _MM_SHUFFLE(0, 0, 3, 2));
+  m = _mm_max_ps (x, s);
+
+  /* m = [max (m[1], m[0])] = [max (max (x[3], x[1]), max (x[2], x[0]))] */
+  s = (__v4sf) _mm_shuffle_epi32 ((__m128i) m, _MM_SHUFFLE(0, 0, 0, 1));
+  m = _mm_max_ps (m, s);
+
+  return m[0];
+}
+
 static inline __v4sf
-init_newton (__v4sf x, double exponent, double c0, double c1, double c2)
+sse_init_newton (__v4sf x, double exponent, double c0, double c1, double c2)
 {
     double norm = exponent*M_LN2/FLT_MANTISSA;
     __v4sf y = _mm_cvtepi32_ps((__m128i)((__v4si)x - splat4i(FLT_ONE)));
@@ -252,10 +278,17 @@ init_newton (__v4sf x, double exponent, double c0, double c1, double c2)
 }
 
 static inline __v4sf
-pow_1_24 (__v4sf x)
+sse_pow_1_24 (__v4sf x)
 {
   __v4sf y, z;
-  y = init_newton (x, -1./12, 0.9976800269, 0.9885126933, 0.5908575383);
+  if (sse_max_component (x) > 1024.0f) {
+    /* for large values, fall back to a slower but more accurate version */
+    return _mm_set_ps (expf (logf (x[3]) * (1.0f / 2.4f)),
+                       expf (logf (x[2]) * (1.0f / 2.4f)),
+                       expf (logf (x[1]) * (1.0f / 2.4f)),
+                       expf (logf (x[0]) * (1.0f / 2.4f)));
+  }
+  y = sse_init_newton (x, -1./12, 0.9976800269, 0.9885126933, 0.5908575383);
   x = _mm_sqrt_ps (x);
   /* newton's method for x^(-1/6) */
   z = splat4f (1.f/6.f) * x;
@@ -265,10 +298,17 @@ pow_1_24 (__v4sf x)
 }
 
 static inline __v4sf
-pow_24 (__v4sf x)
+sse_pow_24 (__v4sf x)
 {
   __v4sf y, z;
-  y = init_newton (x, -1./5, 0.9953189663, 0.9594345146, 0.6742970332);
+  if (sse_max_component (x) > 16.0f) {
+    /* for large values, fall back to a slower but more accurate version */
+    return _mm_set_ps (expf (logf (x[3]) * 2.4f),
+                       expf (logf (x[2]) * 2.4f),
+                       expf (logf (x[1]) * 2.4f),
+                       expf (logf (x[0]) * 2.4f));
+  }
+  y = sse_init_newton (x, -1./5, 0.9953189663, 0.9594345146, 0.6742970332);
   /* newton's method for x^(-1/5) */
   z = splat4f (1.f/5.f) * x;
   y = splat4f (6.f/5.f) * y - z * ((y*y*y)*(y*y*y));
@@ -280,7 +320,10 @@ pow_24 (__v4sf x)
 static inline __v4sf
 linear_to_gamma_2_2_sse2 (__v4sf x)
 {
-  __v4sf curve = pow_1_24 (x) * splat4f (1.055f) - splat4f (0.055f);
+  __v4sf curve = sse_pow_1_24 (x) * splat4f (1.055f) -
+                 splat4f (0.055f                     -
+                          3.0f / (float) (1 << 24));
+                          /* ^ offset the result such that 1 maps to 1 */
   __v4sf line = x * splat4f (12.92f);
   __v4sf mask = _mm_cmpgt_ps (x, splat4f (0.003130804954f));
   return _mm_or_ps (_mm_and_ps (mask, curve), _mm_andnot_ps (mask, line));
@@ -289,15 +332,15 @@ linear_to_gamma_2_2_sse2 (__v4sf x)
 static inline __v4sf
 gamma_2_2_to_linear_sse2 (__v4sf x)
 {
-  __v4sf curve = pow_24 ((x + splat4f (0.055f)) * splat4f (1/1.055f));
+  __v4sf curve = sse_pow_24 ((x + splat4f (0.055f)) * splat4f (1/1.055f));
   __v4sf line = x * splat4f (1/12.92f);
   __v4sf mask = _mm_cmpgt_ps (x, splat4f (0.04045f));
   return _mm_or_ps (_mm_and_ps (mask, curve), _mm_andnot_ps (mask, line));
 }
 
 #define GAMMA_RGBA(func, munge) \
-static long \
-func (const float *src, float *dst, long samples)\
+static inline void \
+func (const Babl *conversion,const float *src, float *dst, long samples)\
 {\
   int i = samples;\
   if (((uintptr_t)src % 16) + ((uintptr_t)dst % 16) == 0)\
@@ -347,12 +390,21 @@ func (const float *src, float *dst, long samples)\
           dst[3] = a;\
         }\
     }\
-  return samples;\
 }
 
 GAMMA_RGBA(conv_rgbaF_linear_rgbaF_gamma, linear_to_gamma_2_2_sse2)
 GAMMA_RGBA(conv_rgbaF_gamma_rgbaF_linear, gamma_2_2_to_linear_sse2)
 
+static void conv_rgbaF_linear_rgbAF_gamma (const Babl  *conversion,
+                                           const float *src,  
+                                           float       *dst, 
+                                           long         samples)
+{
+  float *tmp = alloca (sizeof(float)*4*samples);
+  conv_rgbaF_linear_rgbaF_gamma (conversion, src, tmp, samples);
+  conv_rgbaF_linear_rgbAF_linear (conversion, tmp, dst, samples);
+}
+
 #define YA_APPLY(load, store, convert) \
 { \
   __v4sf yyaa0, yyaa1; \
@@ -371,11 +423,12 @@ GAMMA_RGBA(conv_rgbaF_gamma_rgbaF_linear, gamma_2_2_to_linear_sse2)
   store ((float *)d++, yaya1); \
 }\
 
-static long
-conv_yaF_linear_yaF_gamma (const float *src, float *dst, long samples)
+static void
+conv_yaF_linear_yaF_gamma (const Babl  *conversion,
+                           const float *src,  
+                           float       *dst, 
+                           long         samples)
 {
-  long total = samples;
-
   const __v4sf *s = (const __v4sf*)src;
         __v4sf *d = (__v4sf*)dst;
 
@@ -401,19 +454,19 @@ conv_yaF_linear_yaF_gamma (const float *src, float *dst, long samples)
 
   while (samples--)
     {
-      *dst++ = babl_linear_to_gamma_2_2 (*src++);
+      float y = *src++;
+      *dst++ = linear_to_gamma_2_2_sse2 (splat4f (y))[0];
       *dst++ = *src++;
     }
-
-  return total;
 }
 
 
-static long
-conv_yaF_gamma_yaF_linear (const float *src, float *dst, long samples)
+static void
+conv_yaF_gamma_yaF_linear (const Babl  *conversion,
+                           const float *src,  
+                           float       *dst, 
+                           long         samples)
 {
-  long total = samples;
-
   const __v4sf *s = (const __v4sf*)src;
         __v4sf *d = (__v4sf*)dst;
 
@@ -439,18 +492,18 @@ conv_yaF_gamma_yaF_linear (const float *src, float *dst, long samples)
 
   while (samples--)
     {
-      *dst++ = babl_gamma_2_2_to_linear (*src++);
+      float y = *src++;
+      *dst++ = gamma_2_2_to_linear_sse2 (splat4f (y))[0];
       *dst++ = *src++;
     }
-
-  return total;
 }
 
-static inline long
-conv_yF_linear_yF_gamma (const float *src, float *dst, long samples)
+static inline void
+conv_yF_linear_yF_gamma (const Babl  *conversion,
+                         const float *src,  
+                         float       *dst, 
+                         long         samples)
 {
-  long total = samples;
-
   const __v4sf *s = (const __v4sf*)src;
         __v4sf *d = (__v4sf*)dst;
 
@@ -480,17 +533,17 @@ conv_yF_linear_yF_gamma (const float *src, float *dst, long samples)
 
   while (samples--)
     {
-      *dst++ = babl_linear_to_gamma_2_2 (*src++);
+      float y = *src++;
+      *dst++ = linear_to_gamma_2_2_sse2 (splat4f (y))[0];
     }
-
-  return total;
 }
 
-static inline long
-conv_yF_gamma_yF_linear (const float *src, float *dst, long samples)
+static inline void
+conv_yF_gamma_yF_linear (const Babl  *conversion,
+                         const float *src,  
+                         float       *dst, 
+                         long         samples)
 {
-  long total = samples;
-
   const __v4sf *s = (const __v4sf*)src;
         __v4sf *d = (__v4sf*)dst;
 
@@ -520,24 +573,29 @@ conv_yF_gamma_yF_linear (const float *src, float *dst, long samples)
 
   while (samples--)
     {
-      *dst++ = babl_gamma_2_2_to_linear (*src++);
+      float y = *src++;
+      *dst++ = gamma_2_2_to_linear_sse2 (splat4f (y))[0];
     }
-
-  return total;
 }
 
 
-static long
-conv_rgbF_linear_rgbF_gamma (const float *src, float *dst, long samples)
+static void
+conv_rgbF_linear_rgbF_gamma (const Babl  *conversion,
+                             const float *src,  
+                             float       *dst, 
+                             long         samples)
 {
-  return conv_yF_linear_yF_gamma (src, dst, samples * 3) / 3;
+  conv_yF_linear_yF_gamma (conversion, src, dst, samples * 3);
 }
 
 
-static long
-conv_rgbF_gamma_rgbF_linear (const float *src, float *dst, long samples)
+static void
+conv_rgbF_gamma_rgbF_linear (const Babl  *conversion,
+                             const float *src,  
+                             float       *dst, 
+                             long         samples)
 {
-  return conv_yF_gamma_yF_linear (src, dst, samples * 3) / 3;
+  conv_yF_gamma_yF_linear (conversion, src, dst, samples * 3);
 }
 
 #endif /* defined(USE_SSE2) */
@@ -568,6 +626,14 @@ init (void)
     babl_component ("Ba"),
     babl_component ("A"),
     NULL);
+  const Babl *rgbAF_gamma = babl_format_new (
+    babl_model ("R'aG'aB'aA"),
+    babl_type ("float"),
+    babl_component ("R'a"),
+    babl_component ("G'a"),
+    babl_component ("B'a"),
+    babl_component ("A"),
+    NULL);
   const Babl *rgbaF_gamma = babl_format_new (
     babl_model ("R'G'B'A"),
     babl_type ("float"),
@@ -622,7 +688,19 @@ init (void)
                           "linear",
                           conv_rgbaF_linear_rgbAF_linear,
                           NULL);
+
+      babl_conversion_new(rgbaF_gamma, 
+                          rgbAF_gamma,
+                          "linear",
+                          conv_rgbaF_linear_rgbAF_linear,
+                          NULL);
                           
+      babl_conversion_new(rgbaF_linear, 
+                          rgbAF_gamma,
+                          "linear",
+                          conv_rgbaF_linear_rgbAF_gamma,
+                          NULL);
+
       /* Which of these is faster varies by CPU, and the difference
        * is big enough that it's worthwhile to include both and
        * let them fight it out in the babl benchmarks.
@@ -632,6 +710,12 @@ init (void)
                           "linear",
                           conv_rgbAF_linear_rgbaF_linear_shuffle,
                           NULL);
+      babl_conversion_new(rgbAF_gamma, 
+                          rgbaF_gamma,
+                          "linear",
+                          conv_rgbAF_linear_rgbaF_linear_shuffle,
+                          NULL);
+
       babl_conversion_new(rgbAF_linear, 
                           rgbaF_linear,
                           "linear",
